# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Log all database transactions.

.. note::

   This plugin is for Cylc developers debugging database issues.

Writes an SQL file into the workflow run directory on shutown.

"""

import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path

import sqlparse

from cylc.flow import CYLC_LOG
from cylc.flow.main_loop import (startup, shutdown)


DB_LOG = logging.getLogger(f'{CYLC_LOG}-db')


def _format(sql_string):
    """Pretty print an SQL statement."""
    return '\n'.join(
        sqlparse.format(
            statement,
            reindent_aligned=True,
            use_space_around_operators=True,
            strip_comments=True,
            keyword_case='upper',
            identifier_case='lower',
        )
        for statement in sqlparse.split(sql_string)
    ) + '\n'


def _log(sql_string):
    """Log a SQL string."""
    DB_LOG.info(_format(sql_string))


def _patch_db_connect(db_connect_method):
    """Patch the workflow DAO to configure logging.

    We patch the connect method so that any subsequent re-connections
    are also patched.
    """
    def _inner(*args, **kwargs):
        nonlocal db_connect_method
        conn = db_connect_method(*args, **kwargs)
        conn.set_trace_callback(_log)
        return conn
    return _inner


@startup
async def init(scheduler, state):
    # configure log handler
    DB_LOG.setLevel(logging.INFO)
    handler = RotatingFileHandler(
        str(Path(scheduler.workflow_run_dir, f'{__name__}.sql')),
        maxBytes=1000000,
    )
    state['log_handler'] = handler
    DB_LOG.addHandler(handler)

    # configure the DB manager to log all DB operations
    scheduler.workflow_db_mgr.pri_dao.connect = _patch_db_connect(
        scheduler.workflow_db_mgr.pri_dao.connect
    )


@shutdown
async def stop(scheduler, state):
    handler = state.get('log_handler')
    if handler:
        handler.close()



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
The C3 algorithm is used to linearize multiple inheritance hierarchies
in Python and other languages (MRO = Method Resolution Order). The code
in this doc string is by Samuele Pedroni and Michele Simionato and taken
from the official Python web site, here:
   http://www.python.org/download/releases/2.3/mro/

For cylc, to linearize the runtime inheritance hierarchy, merge()
is unchanged - it works on generic sequences - but mro() had to be
modified to work with lists of names instead of Python base classes.

class __metaclass__(type):
    "All classes are metamagically modified to be nicely printed"
    __repr__ = lambda cls: cls.__name__

class ex_2:
    "Serious order disagreement" #From Guido
    class O: pass
    class X(O): pass
    class Y(O): pass
    class A(X,Y): pass
    class B(Y,X): pass
    try:
        class Z(A,B): pass #creates Z(A,B) in Python 2.2
    except TypeError:
        pass # Z(A,B) cannot be created in Python 2.3

class ex_5:
    "My first example"
    class O: pass
    class F(O): pass
    class E(O): pass
    class D(O): pass
    class C(D,F): pass
    class B(D,E): pass
    class A(B,C): pass

class ex_6:
    "My second example"
    class O: pass
    class F(O): pass
    class E(O): pass
    class D(O): pass
    class C(D,F): pass
    class B(E,D): pass
    class A(B,C): pass

class ex_9:
    "Difference between Python 2.2 MRO and C3" #From Samuele
    class O: pass
    class A(O): pass
    class B(O): pass
    class C(O): pass
    class D(O): pass
    class E(O): pass
    class K1(A,B,C): pass
    class K2(D,B,E): pass
    class K3(D,A): pass
    class Z(K1,K2,K3): pass

def merge(seqs):
    print '\n\nCPL[%s]=%s' % (seqs[0][0],seqs),
    res = []; i=0
    while 1:
      nonemptyseqs=[seq for seq in seqs if seq]
      if not nonemptyseqs: return res
      i+=1; print '\n',i,'round: candidates...',
      for seq in nonemptyseqs: # find merge candidates among seq heads
          cand = seq[0]; print ' ',cand,
          nothead=[s for s in nonemptyseqs if cand in s[1:]]
          if nothead: cand=None #reject candidate
          else: break
      if not cand: raise "Inconsistent hierarchy"
      res.append(cand)
      for seq in nonemptyseqs: # remove cand
          if seq[0] == cand: del seq[0]

def mro(C):
    "Compute the class precedence list (mro) according to C3"
    return merge([[C]]+map(mro,C.__bases__)+[list(C.__bases__)])

def print_mro(C):
    print '\nMRO[%s]=%s' % (C,mro(C))
    print '\nP22 MRO[%s]=%s' % (C,C.mro())

print_mro(ex_9.Z)
"""


from copy import copy


class C3:
    def __init__(self, tree=None):
        if not tree:
            tree = {}
        self.tree = tree

    @staticmethod
    def merge(seqs, label=None):
        # print '\n\nCPL[%s]=%s' % (seqs[0][0],seqs),
        res = []
        while True:
            nonemptyseqs = [seq for seq in seqs if seq]
            if not nonemptyseqs:
                return res
            for seq in nonemptyseqs:  # find merge candidates among seq heads
                cand = seq[0]  # print ' ',cand,
                nothead = [s for s in nonemptyseqs if cand in s[1:]]
                if nothead:
                    cand = None  # reject candidate
                else:
                    break
            if not cand:
                prefix = ""
                if label:
                    prefix = "{0}: ".format(label)
                raise Exception(
                    "ERROR: {0}".format(prefix) +
                    "bad runtime namespace inheritance hierarchy.\n" +
                    "See the cylc documentation on multiple inheritance."
                )
            res.append(cand)
            for seq in nonemptyseqs:  # remove cand
                if seq[0] == cand:
                    del seq[0]

    def mro(self, C):
        """Compute the precedence list (mro) according to C3"""
        # copy() required here for tree to remain unchanged
        return self.merge(
            [[C]] + [self.mro(x) for x in self.tree[C]] + [copy(self.tree[C])],
            label=C)



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Submits job scripts to loadleveler by the ``llsubmit`` command.

.. cylc-scope:: flow.cylc[runtime][<namespace>]

Loadleveler directives can be provided in the flow.cylc file:

.. code-block:: cylc
   :caption: global.cylc

   [platforms]
       [[myplatform]]
           job runner = loadleveler

.. code-block:: cylc
   :caption: flow.cylc

   [runtime]
       [[my_task]]
           platform = myplatform
           execution time limit = PT10M
           [[[directives]]]
               foo = bar
               baz = qux

These are written to the top of the job script like this:

.. code-block:: bash

   #!/bin/bash
   # DIRECTIVES
   # @ foo = bar
   # @ baz = qux
   # @ wall_clock_limit = 660,600
   # @ queue

If ``restart=yes`` is specified as a directive for loadleveler, the job will
automatically trap SIGUSR1, which loadleveler may use to preempt the job. On
trapping SIGUSR1, the job will inform the workflow that it has been vacated by
loadleveler. This will put it back to the submitted state, until it starts
running again.

If :cylc:conf:`execution time limit` is specified, it is used to generate the
``wall_clock_limit`` directive. The setting is assumed to be the soft limit.
The hard limit will be set by adding an extra minute to the soft limit.  Do not
specify the ``wall_clock_limit`` directive explicitly if :cylc:conf:`execution
time limit` is specified. Otherwise, the execution time limit known by the
workflow may be out of sync with what is submitted to the job runner.

.. cylc-scope::

"""

import re

from cylc.flow.id import Tokens


class LoadlevelerHandler():

    """Loadleveler job submission"""

    DIRECTIVE_PREFIX = "# @ "
    KILL_CMD_TMPL = "llcancel '%(job_id)s'"
    POLL_CMD = "llq"
    REC_ID_FROM_SUBMIT_OUT = re.compile(
        r"""\Allsubmit:\sThe\sjob\s"(?P<id>[^"]+)"\s""")
    REC_ERR_FILTERS = [
        re.compile("^llsubmit: Processed command file through Submit Filter:")]
    SUBMIT_CMD_TMPL = "llsubmit '%(job)s'"
    VACATION_SIGNAL = "USR1"

    def format_directives(self, job_conf):
        """Format the job directives for a job file."""
        job_file_path = job_conf['job_file_path']
        directives = job_conf["directives"].__class__()
        tokens = Tokens(job_conf["task_id"], relative=True)
        directives["job_name"] = (
            f'{job_conf["workflow_name"]}.{tokens["task"]}.{tokens["cycle"]}'

        )
        directives["output"] = job_file_path + ".out"
        directives["error"] = job_file_path + ".err"
        if (job_conf["execution_time_limit"] and
                directives.get("wall_clock_limit") is None):
            directives["wall_clock_limit"] = "%d,%d" % (
                job_conf["execution_time_limit"] + 60,
                job_conf["execution_time_limit"])
        for key, value in list(job_conf["directives"].items()):
            directives[key] = value
        lines = []
        for key, value in directives.items():
            if value:
                lines.append("%s%s = %s" % (self.DIRECTIVE_PREFIX, key, value))
            else:
                lines.append("%s%s" % (self.DIRECTIVE_PREFIX, key))
        lines.append("%squeue" % (self.DIRECTIVE_PREFIX))
        return lines

    def filter_submit_output(self, out, err):
        """Filter the stdout/stderr output - suppress process message."""
        new_err = ""
        if err:
            for line in err.splitlines():
                if any(rec.match(line) for rec in self.REC_ERR_FILTERS):
                    continue
                new_err += line + "\n"
        return out, new_err

    @classmethod
    def filter_poll_many_output(cls, out):
        """Return a list of job IDs still in the job runner.

        Drop STEPID from the JOBID.STEPID returned by 'llq'.
        """
        job_ids = []
        for line in out.splitlines():
            try:
                head = line.split(None, 1)[0]
            except IndexError:
                continue
            job_ids.append(".".join(head.split(".")[:2]))
        return job_ids

    def get_vacation_signal(self, job_conf):
        """Return "USR1" if "restart" directive is "yes"."""
        if job_conf["directives"].get("restart") == "yes":
            return self.VACATION_SIGNAL


JOB_RUNNER_HANDLER = LoadlevelerHandler()



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re

from cylc.flow.parsec.exceptions import (
    FileParseError, IncludeFileNotFoundError)


done = []
flist = []

include_re = re.compile(r'\s*%include\s+([\'"]?)(.*?)([\'"]?)\s*$')


def inline(lines, dir_, filename, for_grep=False, viewcfg=None, level=None):
    """Recursive inlining of parsec include-files"""

    global flist
    if level is None:
        # avoid being affected by multiple *different* calls to this function
        flist = [filename]
    else:
        flist.append(filename)
    single = False
    mark = False
    label = False
    if viewcfg:
        mark = viewcfg['mark']
        single = viewcfg['single']
        label = viewcfg['label']
    else:
        viewcfg = {}

    global done

    outf = []
    initial_line_index = 0

    if level is None:
        level = ''
    elif mark:
        level += '!'

    msg = ''

    for line in lines[initial_line_index:]:
        m = include_re.match(line)
        if m:
            q1, match, q2 = m.groups()
            if q1 and (q1 != q2):
                raise FileParseError(
                    "mismatched quotes",
                    line=line,
                    fpath=filename,
                )
            inc = os.path.join(dir_, match)
            if inc not in done:
                if single:
                    done.append(inc)
                if not os.path.isfile(inc):
                    flist.append(inc)
                    raise IncludeFileNotFoundError(flist)
                if for_grep or single or label:
                    outf.append(
                        '#++++ START INLINED INCLUDE FILE ' + match + msg)
                with open(inc, 'r') as handle:
                    finc = [line.rstrip('\n') for line in handle]
                # recursive inclusion
                outf.extend(inline(
                    finc, dir_, inc, for_grep, viewcfg, level))
                if for_grep or single or label:
                    outf.append(
                        '#++++ END INLINED INCLUDE FILE ' + match + msg)

            else:
                outf.append(level + line)
        else:
            # no match
            outf.append(level + line)
    return outf



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""xtrigger function to trigger off of a wall clock time."""

from time import time
from typing import Any, Dict
from cylc.flow.cycling.iso8601 import interval_parse
from cylc.flow.exceptions import WorkflowConfigError


def wall_clock(offset: str = 'PT0S', sequential: bool = True):
    """Trigger at a specific real "wall clock" time relative to the cycle point
    in the graph.

    Clock triggers, unlike other trigger functions, are executed synchronously
    in the main process.

    Args:
        offset:
            ISO 8601 interval to wait after the cycle point is reached in real
            time before triggering. May be negative, in which case it will
            trigger before the real time reaches the cycle point.
        sequential:
            Wall-clock xtriggers are run sequentially by default.
            See :ref:`Sequential Xtriggers` for more details.

    .. versionchanged:: 8.3.0

       The ``sequential`` argument was added.
    """
    # NOTE: This is just a placeholder for the actual implementation.
    # This is only used for validating the signature and for autodocs.
    ...


def _wall_clock(trigger_time: int) -> bool:
    """Actual implementation of wall_clock.

    Return True after the desired wall clock time, or False before.

    Args:
        trigger_time:
            Trigger time as seconds since Unix epoch.
        sequential (bool):
            Used by the workflow to flag corresponding xtriggers as sequential.
    """
    return time() > trigger_time


def validate(args: Dict[str, Any]):
    """Validate and manipulate args parsed from the workflow config.

    NOTE: the xtrigger signature is different to the function signature above

    wall_clock()  # infer zero interval
    wall_clock(PT1H)
    wall_clock(offset=PT1H)

    The offset must be a valid ISO 8601 interval.
    """
    try:
        interval_parse(args["offset"])
    except (ValueError, AttributeError):
        raise WorkflowConfigError(f"Invalid offset: {args['offset']}")



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from copy import deepcopy
import re
from textwrap import dedent
from typing import TYPE_CHECKING, Callable, Iterable, List, Optional

from cylc.flow.context_node import ContextNode
from cylc.flow.parsec.exceptions import (
    ItemNotFoundError,
    NotSingleItemError,
    InvalidConfigError
)
from cylc.flow.parsec.fileparse import parse
from cylc.flow.parsec.util import printcfg
from cylc.flow.parsec.validate import parsec_validate, ParsecValidator as VDR
from cylc.flow.parsec.OrderedDict import OrderedDictWithDefaults
from cylc.flow.parsec.util import itemstr, m_override, replicate, un_many

if TYPE_CHECKING:
    from optparse import Values


class ParsecConfig:
    """Object wrapper for parsec functions."""

    def __init__(
        self,
        spec: 'ConfigNode',
        upgrader: Optional[Callable[[dict, str], None]] = None,
        output_fname: Optional[str] = None,
        tvars: Optional[dict] = None,
        validator: Optional[Callable] = None,
        options: Optional['Values'] = None
    ):
        """Instatiate a parsec config object.

        Args:
            spec: Specification for the config.
            upgrader: An upgrader function, which converts old config items
                to new ones, or returns errors for obsolete items.
            output_fname: Filename to dump parsed config to.
            tvars: Template variables.
            validator: Function checkin that config is valid; defaults to
                ``parsec_validate``.
            options: Command line options.
        """
        self.sparse = OrderedDictWithDefaults()
        self.dense = OrderedDictWithDefaults()
        self.upgrader = upgrader
        self.tvars = tvars
        self.output_fname = output_fname
        self.spec = spec
        if validator is None:
            validator = parsec_validate
        self.validator = validator
        # Get a list of config items which have a private name ``__MANY__``:
        self.manyparents = self._get_namespace_parents()
        self.options = options

    def loadcfg(self, rcfile, title=""):
        """Parse a config file, upgrade or deprecate items if necessary,
        validate it against the spec, and if this is not the first load,
        combine/override with the existing loaded config."""

        sparse = parse(
            rcfile, self.output_fname, self.tvars, opts=self.options)

        if self.upgrader is not None:
            self.upgrader(sparse, title)

        self.validate(sparse)

        if not self.sparse:
            self.sparse = sparse
        else:
            # Already loaded, override with new items.
            replicate(self.sparse, sparse)

    def validate(self, sparse):
        """Validate sparse config against the file spec."""
        return self.validator(sparse, self.spec)

    def expand(self) -> None:
        """Flesh out undefined items with defaults, if any, from the spec."""
        if not self.dense:
            dense = OrderedDictWithDefaults()
            # Populate dict with default values from the spec
            stack = [(dense, self.spec)]
            while stack:
                defs, spec = stack.pop()
                node: ConfigNode
                for node in spec:
                    if not node.is_leaf():
                        if node.name not in defs:
                            defs[node.name] = OrderedDictWithDefaults()
                        stack.append((defs[node.name], node))
                    else:
                        if node.default == ConfigNode.UNSET:
                            if node.vdr and node.vdr.endswith('_LIST'):
                                defs[node.name] = []
                            else:
                                defs[node.name] = None
                        else:
                            defs[node.name] = node.default
            # override defaults with sparse values
            m_override(dense, self.sparse)
            un_many(dense)
            self.dense = dense

    def get(self, keys: Optional[Iterable[str]] = None, sparse: bool = False):
        """Retrieve items or sections, sparse or dense, by list of keys:
        [sec1,sec2,item] =>
            [sec1]
                [[sec2]]
                    item = value

        NOTE: Side effect when sparse is False: the config gets expanded
        (if not previously expanded).
        """
        if sparse:
            cfg = self.sparse
        else:
            self.expand()
            cfg = self.dense

        parents: List[str] = []
        if keys:
            for key in keys:
                try:
                    cfg = cfg[key]
                except (KeyError, TypeError):
                    if (
                        # __MANY__ setting not present:
                        parents in self.manyparents or
                        # setting not present in __MANY__ section:
                        key in self.spec.get(*parents)
                    ):
                        raise ItemNotFoundError(itemstr(parents, key))
                    raise InvalidConfigError(
                        itemstr(parents, key), self.spec.name
                    )
                else:
                    parents.append(key)

        return cfg

    def idump(self, items=None, sparse=False, prefix='',
              oneline=False, none_str='', handle=None):
        """
        items is a list of --item style inputs:
           '[runtime][foo]script'.
        """
        mkeys = []
        null = True
        if items:
            for i in items:
                null = False
                i = i.lstrip('[')
                i = i.rstrip(']')
                j = re.split(r'\]\[*', i)
                mkeys.append(j)
        if null:
            mkeys = [[]]
        self.mdump(mkeys, sparse, prefix, oneline, none_str, handle=handle)

    def mdump(self, mkeys=None, sparse=False, prefix='',
              oneline=False, none_str='', handle=None):
        if oneline:
            items = []
            if mkeys:
                for keys in mkeys:
                    item = self.get(keys, sparse)
                    if isinstance(item, (list, dict)):
                        raise NotSingleItemError(itemstr(keys))
                    if not item:
                        item = none_str or "None"
                    items.append(str(item))
            # TODO - quote items if they contain spaces or comment delimiters?
            print(prefix + ' '.join(items))
        elif mkeys:
            for keys in mkeys:
                self.dump(keys, sparse, prefix, none_str, handle=handle)

    def dump(
        self, keys=None, sparse=False, prefix='', none_str='', handle=None
    ):
        if not keys:
            keys = []
        cfg = self.get(keys, sparse)
        printcfg(
            cfg, prefix=prefix, level=len(keys),
            none_str=none_str, handle=handle
        )

    def _get_namespace_parents(self) -> List[List[str]]:
        """Get a list of the parents of config items which can be user defined.

        For example, where

        .. code-block:: cylc

           [runtime]
               [[my_task]]  # Custom task names.
               [[my_other_task]]

        this function will return ``[runtime]``.
        """
        try:
            return [
                [parent.name for parent in key.parents()][-2::-1]
                # that slice removes the top-level name and reverses
                for _, key in self.spec.walk()
                if key.name == '__MANY__'
            ]
        except AttributeError:
            return []


class ConfigNode(ContextNode):
    """A Cylc configuration schema, section, or setting.

    Attributes:
        vdr:
            The config type (i.e. parsec validator).
        options:
            List of possible options.
            TODO: allow this to be a dict with help info
        default:
            The default value.
        desc:
            A description of the config.
            Note this gets dedented and stripped.
        display_name:
            This is the user-facing name of the config.
            Note the regular ``name`` might be ``__MANY__``.
        meta:
            Another ConfigNode to use as a template for this one.

            This is useful if you want to create a specific instance of
            a generic configuration e.g. ``[elephant]`` from ``[<animal>]``.

            Leaf nodes inherited from the generic config wil have
            ``meta=True``.

    """

    ROOT_NAME_FMT = '{display_name}'
    NODE_NAME_FMT = '[{display_name}]'
    LEAF_NAME_FMT = '{display_name}'
    SEP = ''

    UNSET = '*value unset*'

    __slots__ = ContextNode.__slots__ + (
        'vdr', 'options', 'default', 'desc', 'display_name', 'meta'
    )

    def __init__(
        self,
        name: str,
        vdr: str = VDR.V_STRING,
        default: object = UNSET,
        options: Optional[list] = None,
        desc: Optional[str] = None,
        meta: Optional['ConfigNode'] = None
    ):
        display_name = name
        if name.startswith('<'):
            # if we use <...> as the name, this is a user-definable config
            # * we set the name (for internal use) to __MANY__
            # * we leave the display_name (for external use) unchanged
            name = '__MANY__'

        ContextNode.__init__(self, name)

        if meta:
            # inherit items from the template configuration
            self._children = deepcopy(meta._children)
            if self._children:
                for child in self._children.values():
                    # record that these configurations have been inherited
                    # (this is used to prevent documenting settings twice)
                    child.meta = True

        self.display_name = display_name
        self.vdr = vdr
        self.default = default
        self.options = options
        self.desc = dedent(desc).strip() if desc else None
        self.meta = meta

    def __repr__(self):
        parents = list(self.parents())
        itr = list(reversed(list(parents))) + [self]
        if len(parents) == 1 and self.is_leaf():
            itr.insert(1, '|')
        return ''.join(map(str, itr))



#!/usr/bin/env python3

# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""cylc ext-trigger [OPTIONS] ARGS

Report an external event message to a scheduler.

External triggers allow any program to send a message to the Cylc scheduler.
Cylc can use this message as a signal that an external prerequisite has been
satisfied and trigger the task accordingly.

The WORKFLOW argument should be unique to each external trigger event. When
an incoming message satisfies a task's external trigger the message TRIGGER_ID
is broadcast to all downstream tasks in the cycle point as
``$CYLC_EXT_TRIGGER_ID``.  Tasks can use ``$CYLC_EXT_TRIGGER_ID``, for example,
to identify a new data file that the external triggering system is responding
to.

Use the retry options in case the target workflow is down or out of contact.

Note: to manually trigger a task use 'cylc trigger', not this command.
"""

from time import sleep
from typing import TYPE_CHECKING

from cylc.flow import LOG
from cylc.flow.exceptions import CylcError, ClientError
from cylc.flow.id_cli import parse_id
from cylc.flow.network.client_factory import get_client
from cylc.flow.option_parsers import (
    WORKFLOW_ID_ARG_DOC,
    CylcOptionParser as COP,
)
from cylc.flow.terminal import cli_function

if TYPE_CHECKING:
    from optparse import Values


MAX_N_TRIES = 5
RETRY_INTVL_SECS = 10.0

MSG_SEND_FAILED = "Send message: try %s of %s failed"
MSG_SEND_RETRY = "Retrying in %s seconds, timeout is %s"
MSG_SEND_SUCCEED = "Send message: try %s of %s succeeded"

MUTATION = '''
mutation (
  $wFlows: [WorkflowID]!,
  $eventMsg: String!,
  $eventId: String!
) {
  extTrigger (
    workflows: $wFlows,
    message: $eventMsg,
    id: $eventId
  ) {
    result
  }
}
'''


def get_option_parser() -> COP:
    parser = COP(
        __doc__,
        comms=True,
        argdoc=[
            WORKFLOW_ID_ARG_DOC,
            ("MSG", "External trigger message"),
            ("TRIGGER_ID", "Unique trigger ID"),
        ],
    )

    parser.add_option(
        "--max-tries",
        help=r"Maximum number of send attempts (default: %default).",
        metavar="INT",
        action="store",
        default=MAX_N_TRIES,
        dest="max_n_tries"
    )

    parser.add_option(
        "--retry-interval",
        help=r"Delay in seconds before retrying (default: %default).",
        metavar="SEC",
        action="store",
        default=RETRY_INTVL_SECS,
        dest="retry_intvl_secs"
    )

    return parser


@cli_function(get_option_parser)
def main(
    parser: COP,
    options: 'Values',
    workflow_id: str,
    event_msg: str,
    event_id: str
) -> None:
    workflow_id, *_ = parse_id(
        workflow_id,
        constraint='workflows',
    )
    LOG.info(
        'Send to workflow %s: "%s" (%s)', workflow_id, event_msg, event_id
    )
    pclient = get_client(workflow_id, timeout=options.comms_timeout)

    max_n_tries = int(options.max_n_tries)
    retry_intvl_secs = float(options.retry_intvl_secs)

    mutation_kwargs = {
        'request_string': MUTATION,
        'variables': {
            'wFlows': [workflow_id],
            'eventMsg': event_msg,
            'eventId': event_id,
        }
    }

    for i_try in range(max_n_tries):
        try:
            pclient('graphql', mutation_kwargs)
        except ClientError as exc:
            LOG.exception(exc)
            LOG.info(MSG_SEND_FAILED, i_try + 1, max_n_tries)
            if i_try == max_n_tries - 1:  # final attempt
                raise CylcError('send failed')
            LOG.info(MSG_SEND_RETRY, retry_intvl_secs, options.comms_timeout)
            sleep(retry_intvl_secs)
        else:
            if i_try > 0:
                LOG.info(MSG_SEND_SUCCEED, i_try + 1, max_n_tries)
            break



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""cylc support for the EmPy template processor

Importing code should catch ImportError in case EmPy is not installed.
"""

from io import StringIO
import em
import os
import typing as t

from cylc.flow.parsec.exceptions import EmPyError
from cylc.flow.parsec.fileparse import get_cylc_env_vars


def empyprocess(
    _fpath: str,
    flines: t.List[str],
    dir_: str,
    template_vars: t.Optional[t.Dict[str, t.Any]] = None,
) -> t.List[str]:
    """Pass configure file through EmPy processor.

    Args:
        _fpath:
            The path to the root template file (i.e. the flow.cylc file)
        flines:
            List of template lines to process.
        dir_:
            The path to the configuration directory.
        template_vars:
            Dictionary of template variables.

    """

    cwd = os.getcwd()

    os.chdir(dir_)
    ftempl = StringIO('\n'.join(flines))
    xtempl = StringIO()
    interpreter = em.Interpreter(output=em.UncloseableFile(xtempl))

    # Add `CYLC_` environment variables to the global namespace.
    interpreter.updateGlobals(
        get_cylc_env_vars()
    )

    try:
        interpreter.file(ftempl, '<template>', template_vars)
    except Exception as exc:
        lineno = interpreter.contexts[-1].identify()[1]
        raise EmPyError(
            str(exc),
            lines={'<template>': flines[max(lineno - 4, 0): lineno]},
        )
    finally:
        interpreter.shutdown()
        xworkflow = xtempl.getvalue()
        os.chdir(cwd)
        ftempl.close()
        xtempl.close()

    flow_config = []
    for line in xworkflow.splitlines():
        # EmPy leaves blank lines where source lines contain
        # only EmPy code; this matters if line continuation
        # markers are involved, so we remove blank lines here.
        if not line.strip():
            continue
            # restoring newlines here is only necessary for display by
        # the cylc view command:
        # ##flow_config.append(line + '\n')
        flow_config.append(line)

    return flow_config



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Workflow status constants."""

from enum import Enum
from typing import Tuple, TYPE_CHECKING

from cylc.flow.wallclock import get_time_string_from_unix_time as time2str

if TYPE_CHECKING:
    from optparse import Values
    from cylc.flow.scheduler import Scheduler

# Keys for identify API call
KEY_GROUP = "group"
KEY_META = "meta"
KEY_NAME = "name"
KEY_OWNER = "owner"
KEY_STATES = "states"
KEY_TASKS_BY_STATE = "tasks-by-state"
KEY_UPDATE_TIME = "update-time"
KEY_VERSION = "version"

# Status message strings
WORKFLOW_STATUS_RUNNING_TO_STOP = "running to stop at %s"
WORKFLOW_STATUS_RUNNING_TO_HOLD = "running to hold at %s"

# TODO - the workflow status should be a property of the Scheduler instance
#        rather than derived from scheduler state?


class WorkflowStatus(Enum):
    """The possible statuses of a workflow."""

    PAUSED = "paused"
    """Workflow will not submit any new jobs."""

    RUNNING = "running"
    """Workflow is running as normal."""

    STOPPING = "stopping"
    """Workflow is in the process of shutting down."""

    STOPPED = "stopped"
    """Workflow is not running."""


class StopMode(Enum):
    """The possible modes of a workflow shutdown"""

    AUTO = 'AUTOMATIC'
    """Workflow has reached a state where it can automatically stop"""

    AUTO_ON_TASK_FAILURE = 'AUTOMATIC(ON-TASK-FAILURE)'
    """A task has failed and ``--abort-if-any-task-fails`` was used."""

    REQUEST_CLEAN = 'REQUEST(CLEAN)'
    """External shutdown request, will wait for active jobs to complete."""

    REQUEST_KILL = 'REQUEST(KILL)'
    """External shutdown request, will wait for active jobs to be killed."""

    REQUEST_NOW = 'REQUEST(NOW)'
    """External shutdown request, will wait for event handlers to complete."""

    REQUEST_NOW_NOW = 'REQUEST(NOW-NOW)'
    """External immediate shutdown request."""

    def explain(self) -> str:
        """Return an explanation of what a workflow in this state is doing.

        This is used in the workflow status message if the workflow is stopping
        to clarify what the scheduler is doing.
        """
        if self in (self.AUTO, self.REQUEST_CLEAN):  # type: ignore
            return 'waiting for active jobs to complete'
        if self == self.REQUEST_KILL:  # type: ignore
            return 'killing active jobs'
        if self in (  # type: ignore
            self.AUTO_ON_TASK_FAILURE,  # type: ignore
            self.REQUEST_NOW,
            self.REQUEST_NOW_NOW,
        ):
            return 'shutting down'
        return ''

    def describe(self) -> str:
        """Return a user-friendly description of this state.

        This is used in the schema to convey what the different stop modes do.
        """
        if self == self.AUTO:  # type: ignore
            return 'Wait until workflow has completed.'
        if self == self.AUTO_ON_TASK_FAILURE:  # type: ignore
            return 'Wait until the first task fails.'
        if self == self.REQUEST_CLEAN:  # type: ignore
            return (
                'Regular shutdown:\n'
                '* Wait for all active jobs to complete.\n'
                '* Run workflow event handlers and wait for them to complete.'
            )
        if self == self.REQUEST_KILL:  # type: ignore
            return (
                'Kill shutdown:\n'
                '* Wait for all active jobs to be killed.\n'
                '* Run workflow event handlers and wait for them to complete.'
            )
        if self == self.REQUEST_NOW:  # type: ignore
            return (
                'Immediate shutdown\n'
                "* Don't kill submitted or running jobs.\n"
                '* Run workflow event handlers and wait for them to complete.'
            )
        if self == self.REQUEST_NOW_NOW:  # type: ignore
            return (
                'Immediate shutdown\n'
                "* Don't kill submitted or running jobs.\n"
                "* Don't run event handlers."
            )
        raise KeyError(f'No description for {self}.')


class AutoRestartMode(Enum):
    """The possible modes of a workflow auto-restart."""

    RESTART_NORMAL = 'stop and restart'
    """Workflow will stop immediately and attempt to restart."""

    FORCE_STOP = 'stop'
    """Workflow will stop immediately but *not* attempt to restart."""


def get_workflow_status(schd: 'Scheduler') -> Tuple[str, str]:
    """Return the status of the provided workflow.

    This should be a short, concise description of the workflow state.

    Args:
        schd: The running workflow

    Returns:
        tuple - (state, state_msg)

        state:
            The WorkflowState.
        state_msg:
            Text describing the current state (may be an empty string).

    """
    status = WorkflowStatus.RUNNING
    status_msg = ''

    if schd.stop_mode is not None:
        status = WorkflowStatus.STOPPING
        status_msg = f'stopping: {schd.stop_mode.explain()}'
    elif schd.reload_pending:
        status = WorkflowStatus.PAUSED
        status_msg = f'reloading: {schd.reload_pending}'
    elif schd.is_stalled:
        status_msg = 'stalled'
    elif schd.is_paused:
        status = WorkflowStatus.PAUSED
        status_msg = 'paused'
    elif schd.pool.hold_point:
        status_msg = (
            WORKFLOW_STATUS_RUNNING_TO_HOLD %
            schd.pool.hold_point)
    elif schd.pool.stop_point:
        status_msg = (
            WORKFLOW_STATUS_RUNNING_TO_STOP %
            schd.pool.stop_point)
    elif schd.stop_clock_time is not None:
        status_msg = (
            WORKFLOW_STATUS_RUNNING_TO_STOP %
            time2str(schd.stop_clock_time))
    elif schd.pool.stop_task_id:
        status_msg = (
            WORKFLOW_STATUS_RUNNING_TO_STOP %
            schd.pool.stop_task_id)
    elif schd.config and schd.config.final_point:
        status_msg = (
            WORKFLOW_STATUS_RUNNING_TO_STOP %
            schd.config.final_point)
    else:
        # fallback - running indefinitely
        status_msg = 'running'

    return (status.value, status_msg)


class RunMode:
    """The possible run modes of a workflow."""

    LIVE = 'live'
    """Workflow will run normally."""

    SIMULATION = 'simulation'
    """Workflow will run in simulation mode."""

    DUMMY = 'dummy'
    """Workflow will run in dummy mode."""

    @staticmethod
    def get(options: 'Values') -> str:
        """Return the run mode from the options."""
        return getattr(options, 'run_mode', None) or RunMode.LIVE



# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Module for parsing cylc graph strings."""

import re
import contextlib

from typing import (
    Set,
    Dict,
    List,
    Tuple,
    Optional,
    Union
)

import cylc.flow.flags
from cylc.flow.exceptions import GraphParseError
from cylc.flow.param_expand import GraphExpander
from cylc.flow.task_id import TaskID
from cylc.flow.task_trigger import TaskTrigger
from cylc.flow.task_outputs import (
    TASK_OUTPUT_EXPIRED,
    TASK_OUTPUT_SUCCEEDED,
    TASK_OUTPUT_STARTED,
    TASK_OUTPUT_FAILED,
    TASK_OUTPUT_FINISHED,
    TASK_OUTPUT_SUBMITTED,
    TASK_OUTPUT_SUBMIT_FAILED
)
from cylc.flow.task_qualifiers import (
    QUAL_FAM_EXPIRE_ALL,
    QUAL_FAM_EXPIRE_ANY,
    QUAL_FAM_SUCCEED_ALL,
    QUAL_FAM_SUCCEED_ANY,
    QUAL_FAM_FAIL_ALL,
    QUAL_FAM_FAIL_ANY,
    QUAL_FAM_FINISH_ALL,
    QUAL_FAM_FINISH_ANY,
    QUAL_FAM_START_ALL,
    QUAL_FAM_START_ANY,
    QUAL_FAM_SUBMIT_ALL,
    QUAL_FAM_SUBMIT_ANY,
    QUAL_FAM_SUBMIT_FAIL_ALL,
    QUAL_FAM_SUBMIT_FAIL_ANY,
)


class Replacement:
    """A class to remember match group information in re.sub() calls"""
    def __init__(self, replacement):
        self.replacement = replacement
        self.substitutions = []
        self.match_groups = []

    def __call__(self, match):
        matched = match.group(0)
        replaced = match.expand(self.replacement)
        self.substitutions.append((matched, replaced))
        self.match_groups.append(match.groups())
        return replaced


class GraphParser:
    """Class for extracting dependency information from cylc graph strings.

    For each task in the graph string, results are stored as:
        self.triggers[task_name][expression] = ([expr_task_names], suicide)
        self.original[task_name][expression] = original_expression

    (original_expression is separated out to allow comparison of triggers
    from different equivalent expressions, e.g. family vs member).

    This is currently intended to process a single multi-line graph string
    (i.e. the content of a single graph section). But it could be extended to
    store dependencies for the whole workflow (call parse_graph multiple times
    and key results by graph section).

    The general form of a dependency is "LHS => RHS", where:
        * On the left, an EXPRESSION of nodes involving parentheses, and
          logical operators '&' (AND), and '|' (OR).
        * On the right, an EXPRESSION of nodes NOT involving '|'
        * Node names may be parameterized (any number of parameters):
            NODE<i,j,k>
            NODE<i=0,j,k>  # specific parameter value
            NODE<i-1,j,k>  # offset parameter value
        * A parameterized qualified node name looks like this:
            NODE(<PARAMS>)([CYCLE-POINT-OFFSET])(:QUALIFIER)
        * The default trigger type is ':succeeded'.
        * A remote workflow qualified node name looks like this:
            NODE(<REMOTE-WORKFLOW-QUALIFIER>)(:QUALIFIER)
        * Outputs (boo:x) are ignored as triggers on the RHS to allow chaining:
            "foo => bar:x => baz & qux"
    """

    CYLC7_COMPAT = "CYLC 7 BACK-COMPAT"

    OP_AND = '&'
    OP_OR = '|'
    OP_AND_ERR = '&&'
    OP_OR_ERR = '||'
    SUICIDE = '!'
    OPTIONAL = '?'
    QUALIFIER = ':'
    ARROW = '=>'
    XTRIG = '@'
    CONTINUATION_STRS = (ARROW, OP_AND, OP_OR)
    BAD_STRS = (OP_AND_ERR, OP_OR_ERR)

    # Map family trigger type to (member-trigger, any/all), for use in
    # expanding family trigger expressions to member trigger expressions.
    # - "FAM:succeed-all => g" means "f1:succeed & f2:succeed => g"
    # - "FAM:fail-any => g" means "f1:fail | f2:fail => g".
    # E.g. QUAL_FAM_START_ALL: (TASK_OUTPUT_STARTED, True) simply maps
    #   "FAM:start-all" to "MEMBER:started" and "-all" (all members).
    fam_to_mem_trigger_map: Dict[str, Tuple[str, bool]] = {
        QUAL_FAM_EXPIRE_ALL: (TASK_OUTPUT_EXPIRED, True),
        QUAL_FAM_EXPIRE_ANY: (TASK_OUTPUT_EXPIRED, False),
        QUAL_FAM_START_ALL: (TASK_OUTPUT_STARTED, True),
        QUAL_FAM_START_ANY: (TASK_OUTPUT_STARTED, False),
        QUAL_FAM_SUCCEED_ALL: (TASK_OUTPUT_SUCCEEDED, True),
        QUAL_FAM_SUCCEED_ANY: (TASK_OUTPUT_SUCCEEDED, False),
        QUAL_FAM_FAIL_ALL: (TASK_OUTPUT_FAILED, True),
        QUAL_FAM_FAIL_ANY: (TASK_OUTPUT_FAILED, False),
        QUAL_FAM_SUBMIT_ALL: (TASK_OUTPUT_SUBMITTED, True),
        QUAL_FAM_SUBMIT_ANY: (TASK_OUTPUT_SUBMITTED, False),
        QUAL_FAM_SUBMIT_FAIL_ALL: (TASK_OUTPUT_SUBMIT_FAILED, True),
        QUAL_FAM_SUBMIT_FAIL_ANY: (TASK_OUTPUT_SUBMITTED, False),
        QUAL_FAM_FINISH_ALL: (TASK_OUTPUT_FINISHED, True),
        QUAL_FAM_FINISH_ANY: (TASK_OUTPUT_FINISHED, False),
    }

    # Map family pseudo triggers to affected member outputs.
    fam_to_mem_output_map: Dict[str, List[str]] = {
        QUAL_FAM_EXPIRE_ANY: [TASK_OUTPUT_EXPIRED],
        QUAL_FAM_EXPIRE_ALL: [TASK_OUTPUT_EXPIRED],
        QUAL_FAM_START_ANY: [TASK_OUTPUT_STARTED],
        QUAL_FAM_START_ALL: [TASK_OUTPUT_STARTED],
        QUAL_FAM_SUCCEED_ANY: [TASK_OUTPUT_SUCCEEDED],
        QUAL_FAM_SUCCEED_ALL: [TASK_OUTPUT_SUCCEEDED],
        QUAL_FAM_FAIL_ANY: [TASK_OUTPUT_FAILED],
        QUAL_FAM_FAIL_ALL: [TASK_OUTPUT_FAILED],
        QUAL_FAM_SUBMIT_ANY: [TASK_OUTPUT_SUBMITTED],
        QUAL_FAM_SUBMIT_ALL: [TASK_OUTPUT_SUBMITTED],
        QUAL_FAM_SUBMIT_FAIL_ANY: [TASK_OUTPUT_SUBMIT_FAILED],
        QUAL_FAM_SUBMIT_FAIL_ALL: [TASK_OUTPUT_SUBMIT_FAILED],
        QUAL_FAM_FINISH_ANY: [TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED],
        QUAL_FAM_FINISH_ALL: [TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED]
    }

    _RE_SUICIDE = r'(?:!)?'
    _RE_NODE = _RE_SUICIDE + TaskID.NAME_RE
    _RE_NODE_OR_XTRIG = r'(?:[!@])?' + TaskID.NAME_RE
    _RE_PARAMS = r'<[\w,=\-+]+>'
    _RE_OFFSET = r'\[[\w\-\+\^:]+\]'
    _RE_QUAL = QUALIFIER + r'[\w\-]+'  # task or fam trigger
    _RE_OPT = r'\??'  # optional output indicator

    REC_QUAL = re.compile(_RE_QUAL)

    # Match if there are any spaces which could lead to graph problems
    REC_GRAPH_BAD_SPACES_LINE = re.compile(
        TaskID.NAME_RE +
        rf"""
        (?<![\-+](?=\s*[0-9]))  # allow spaces after -+ if numbers follow
        (?!\s*[\-+]\s*[0-9])    # allow spaces before/after -+ if nums follow
        \s+                     # do not allow 'task<space>task'
        {TaskID.NAME_SUFFIX_RE}
        """,
        re.X
    )

    # Match @xtriggers.
    REC_XTRIG = re.compile(r'@[\w\-+%]+')

    # Match fully qualified parameterized single nodes.
    REC_NODE_FULL = re.compile(
        rf"""
        {_RE_SUICIDE}
        (?:(?:{TaskID.NAME_RE}             # node name
        (?:{_RE_PARAMS})?|{_RE_PARAMS}))+  # allow task<param> to repeat
        (?:{_RE_OFFSET})?                  # cycle point offset
        (?:{_RE_QUAL})?                    # qualifier
        (?:{_RE_OPT})?                     # optional output indicator
        """,
        re.X
    )

    # Extract node or xtrigger from LHS expressions after param expansion.
    REC_NODES = re.compile(
        rf"""
        ({_RE_NODE_OR_XTRIG})  # node name
        ({_RE_OFFSET})?        # cycle point offset
        ({_RE_QUAL})?          # trigger qualifier
        ({_RE_OPT})?           # optional output indicator
        """,
        re.X
    )

    REC_COMMENT = re.compile('#.*$')

    # Detect presence of expansion parameters in a graph line.
    REC_PARAMS = re.compile(_RE_PARAMS)

    # Detect and extract workflow state polling task info.
    REC_WORKFLOW_STATE = re.compile(
        rf"""
        ({TaskID.NAME_RE})
        (<([\w.\-/]+)::({TaskID.NAME_RE})
        ({_RE_QUAL})?>)
        """,
        re.X
    )

    # Remove out-of-range nodes
    # <TASK_NAME_PART> : [^\s&\|] # i.e. sequence of not <AND|OR|SPACE>
    # <REMOVE_TOKEN>   : <TASK_NAME_PART> <_REMOVE>
    #                    <TASK_NAME_PART>?
    _TASK_NAME_PART = rf'[^\s{OP_AND}{OP_OR}]'
    _REMOVE_TOKEN = (
        rf"{_TASK_NAME_PART}+{str(GraphExpander._REMOVE)}?{_TASK_NAME_PART}+"
    )

    REC_NODE_OUT_OF_RANGE = re.compile(
        rf"""
        (                                     #
            ^{_REMOVE_TOKEN}[{OP_AND}{OP_OR}] # ^<REMOVE> <AND|OR>
            |                                 #
            [{OP_AND}{OP_OR}]{_REMOVE_TOKEN}  # <AND|OR> <REMOVE>
            |                                 #
            ^{_REMOVE_TOKEN}$                 # ^<REMOVE>$
        )                                     #
        """,
        re.X
    )

    REC_RHS_NODE = re.compile(
        rf"""
        (!)?           # suicide mark
        ({_RE_NODE})   # node name
        ({_RE_QUAL})?  # trigger qualifier
        ({_RE_OPT})?   # optional output indicator
        """,
        re.X
    )

    def __init__(
        self,
        family_map: Optional[Dict[str, List[str]]] = None,
        parameters: Optional[Dict] = None,
        task_output_opt:
            Optional[Dict[Tuple[str, str], Tuple[bool, bool, bool]]] = None
    ) -> None:
        """Initialize the graph string parser.

        Args:
            family_map:
                {family_name: [family member task names]}
            parameters:
                task parameters for expansion here
            task_output_opt:
                {(name, output): (is-optional, is-opt-default, is-fixed)}
                passed in to allow checking across multiple graph strings

        """
        self.family_map = family_map or {}
        self.parameters = parameters
        self.triggers: Dict = {}
        self.original: Dict = {}
        self.workflow_state_polling_tasks: Dict = {}

        # Record task outputs as optional or required:
        #   {(name, output): (is_optional, is_member)}
        # (Need to compare across separately-parsed graph strings.)
        if task_output_opt:
            self.task_output_opt = task_output_opt
        else:
            self.task_output_opt = {}

    def parse_graph(self, graph_string: str) -> None:
        """Parse the graph string for a single graph section.

        (Assumes any general line-continuation markers have been processed).
           1. Strip comments, whitespace, and blank lines.
              (all whitespace is removed up front so we don't have to consider
              it in regexes and strip it from matched elements)
           2. Join incomplete lines starting or ending with '=>'.
           3. Replicate and expand any parameterized lines.
           4. Split and process by pairs "left-expression => right-node":
              i. Replace families with members (any or all semantics).
             ii. Record parsed dependency information for each right-side node.
        """
        # Strip comments, whitespace, and blank lines.
        non_blank_lines = []
        bad_lines = []
        for line in graph_string.split('\n'):
            modified_line = self.__class__.REC_COMMENT.sub('', line)

            # Ignore empty lines
            if not modified_line or modified_line.isspace():
                continue

            # Catch simple bad lines that would be accepted once
            # spaces are removed, e.g. 'foo bar => baz'
            if self.REC_GRAPH_BAD_SPACES_LINE.search(modified_line):
                bad_lines.append(line)
                continue

            # Apparently this is the fastest way to strip all whitespace!:
            modified_line = "".join(modified_line.split())
            non_blank_lines.append(modified_line)

        # Check if there were problem lines and abort
        if bad_lines:
            self._report_invalid_lines(bad_lines)

        # Join incomplete lines (beginning or ending with an arrow).
        full_lines = []
        part_lines = []
        for i, _ in enumerate(non_blank_lines):
            this_line = non_blank_lines[i]
            for seq in self.CONTINUATION_STRS:
                if i == 0 and this_line.startswith(seq):
                    # First line can't start with an arrow.
                    raise GraphParseError(f"Leading {seq}: {this_line}")
            try:
                next_line = non_blank_lines[i + 1]
            except IndexError:
                next_line = ''
                for seq in self.CONTINUATION_STRS:
                    if this_line.endswith(seq):
                        # Last line can't end with an arrow, & or |.
                        raise GraphParseError(
                            f"Dangling {seq}:"
                            f"{this_line}"
                        )
            part_lines.append(this_line)

            # Check that a continuation sequence doesn't end this line and
            # begin the next:
            if (
                this_line.endswith(self.CONTINUATION_STRS) and
                next_line.startswith(self.CONTINUATION_STRS)
            ):
                raise GraphParseError(
                    'Consecutive lines end and start with continuation '
                    'characters:\n'
                    f'{this_line}\n'
                    f'{next_line}'
                )

            # Check that line ends with a valid continuation sequence:
            if (any(
                this_line.endswith(seq) or next_line.startswith(seq) for
                seq in self.CONTINUATION_STRS
            ) and not (any(
                this_line.endswith(seq) or next_line.startswith(seq) for
                seq in self.BAD_STRS
            ))):
                continue

            full_line = ''.join(part_lines)

            # Record inter-workflow dependence and remove the marker notation.
            # ("foo<WORKFLOW::TASK:fail> => bar" becomes:fail "foo => bar").
            repl = Replacement('\\1')
            full_line = self.__class__.REC_WORKFLOW_STATE.sub(repl, full_line)
            for item in repl.match_groups:
                l_task, r_all, r_workflow, r_task, r_status = item
                if r_status is not None:
                    r_status = TaskTrigger.standardise_name(
                        r_status.strip(self.__class__.QUALIFIER)
                    )
                self.workflow_state_polling_tasks[l_task] = (
                    r_workflow, r_task, r_status, r_all
                )

            full_lines.append(full_line)
            part_lines = []

        # Check for double-char conditional operators (a common mistake),
        # and bad node syntax (order of qualifiers).
        bad_lines = []
        for line in full_lines:
            if self.__class__.OP_AND_ERR in line:
                raise GraphParseError(
                    "The graph AND operator is "
                    f"'{self.__class__.OP_AND}': {line}")
            if self.__class__.OP_OR_ERR in line:
                raise GraphParseError(
                    "The graph OR operator is "
                    f"'{self.__class__.OP_OR}': {line}")
            # Check node syntax. First drop all non-node characters.
            node_str = line
            for spec in [
                self.__class__.ARROW,
                self.__class__.OP_OR,
                self.__class__.OP_AND,
                self.__class__.SUICIDE,
                '(',
                ')',
            ]:
                node_str = node_str.replace(spec, ' ')
            # Drop all valid @xtriggers, longest first to avoid sub-strings.
            nodes = self.__class__.REC_XTRIG.findall(node_str)
            nodes.sort(key=len, reverse=True)
            for node in nodes:
                node_str = node_str.replace(node, '')
            # Then drop all valid nodes, longest first to avoid sub-strings.
            bad_lines = [node_str for node in node_str.split()
                         if self.__class__.REC_NODE_FULL.sub('', node, 1)]
        if bad_lines:
            self._report_invalid_lines(bad_lines)

        # Expand parameterized lines (or detect undefined parameters).
        line_set = set()
        graph_expander = GraphExpander(self.parameters)
        for line in full_lines:
            if not self.__class__.REC_PARAMS.search(line):
                line_set.add(line)
                continue
            for line_ in graph_expander.expand(line):
                line_set.add(line_)

        # Process chains of dependencies as pairs: left => right.
        # Parameterization can duplicate some dependencies, so use a set.
        pairs: Set[Tuple[Optional[str], str]] = set()
        for line in line_set:
            chain = []
            # "foo => bar => baz" becomes [foo, bar, baz]
            # "foo => bar_-32768 => baz" becomes [foo]
            # "foo_-32768 => bar" becomes []
            for node in line.split(self.__class__.ARROW):
                # This can happen, e.g. "foo => => bar" produces
                # "foo, '', bar", so we add so that later it raises
                # an error
                if node == '':
                    chain.append(node)
                    continue
                node = self.REC_NODE_OUT_OF_RANGE.sub('', node)
                if node == '':
                    # For "foo => bar<err> => baz", stop at "bar<err>"
                    break
                else:
                    chain.append(node)

            if not chain:
                continue

            for item in self.__class__.REC_NODES.findall(chain[0]):
                # Auto-trigger lone nodes and initial nodes in a chain.
                if not item[0].startswith(self.__class__.XTRIG):
                    pairs.add((None, ''.join(item)))

            for i in range(0, len(chain) - 1):
                pairs.add((chain[i], chain[i + 1]))

        # Get a set of RH nodes which are not at the LH of another pair:
        pairs_dict = dict(pairs)
        terminals = set(pairs_dict.values()).difference(pairs_dict.keys())

        for pair in pairs:
            self._proc_dep_pair(pair, terminals)

    @classmethod
    def _report_invalid_lines(cls, lines: List[str]) -> None:
        """Raise GraphParseError in a consistent format when there are
        lines with bad syntax.

        The list of bad lines are inserted into the error message to show
        exactly what lines have problems. The correct syntax of graph lines
        is displayed to direct people on the correct path.

        Keyword Arguments:
        lines -- a list of bad graph lines to be reported

        Raises:
        GraphParseError -- always. This is the sole purpose of this method
        """
        raise GraphParseError(
            "Bad graph node format:\n"
            "  " + "\n  ".join(lines) + "\n"
            "Correct format is:\n"
            " @XTRIG or "
            " NAME(<PARAMS>)([CYCLE-POINT-OFFSET])(:QUALIFIER)(?)\n"
            " {NAME(<PARAMS>) can also be: "
            "<PARAMS>NAME or NAME<PARAMS>NAME_CONTINUED}\n"
            " or\n"
            " NAME(<REMOTE-WORKFLOW-QUALIFIER>)(:QUALIFIER)")

    def _proc_dep_pair(
        self,
        pair: Tuple[Optional[str], str],
        terminals: Set[str],
    ) -> None:
        """Process a single dependency pair 'left => right'.

        A node is an xtrigger, or a task or a family name.
        A qualified name is NAME([CYCLE-POINT-OFFSET])(:QUALIFIER).

        Args:
            pair:
                'left' can be a logical expression of qualified node names.
                'left' can be None, when triggering a left-side or lone node.
                'left' can be "", if null task name in graph error (a => => b).
                'right' can be one or more node names joined by AND.
                'right' can't be None or "".
            terminals:
                Nodes which are _only_ on the RH end of chains.
        """
        left, right = pair
        # Raise error for right-hand-side OR operators.
        if self.__class__.OP_OR in right:
            raise GraphParseError(f"Illegal OR on right side: {right}")

        # Raise error if suicide triggers on the left of the trigger.
        if left and self.__class__.SUICIDE in left:
            raise GraphParseError(
                "Suicide markers must be"
                f" on the right of a trigger: {left}")

        # Check that parentheses match.
        mismatch_msg = 'Mismatched parentheses in: "{}"'
        if left and left.count("(") != left.count(")"):
            raise GraphParseError(mismatch_msg.format(left))
        if right.count("(") != right.count(")"):
            raise GraphParseError(mismatch_msg.format(right))

        # Raise error for cycle point offsets at the end of chains
        if '[' in right:
            if left and (right in terminals):
                # This right hand side is at the end of a chain:
                raise GraphParseError(
                    'Invalid cycle point offsets only on right hand '
                    'side of a dependency (must be on left hand side):'
                    f' {left} => {right}')
            else:
                # This RHS is also a LHS in a chain:
                return

        # Split right side on AND.
        rights = right.split(self.__class__.OP_AND)
        if '' in rights or right and not all(rights):
            raise GraphParseError(
                f"Null task name in graph: {left} => {right}")

        lefts: Union[List[str], List[Optional[str]]]
        if not left or (self.__class__.OP_OR in left or '(' in left):
            # Treat conditional or parenthesised expressions as a single entity
            # Can get [None] or [""] here
            lefts = [left]
        else:
            # Split non-conditional left-side expressions on AND.
            # Can get [""] here too
            lefts = left.split(self.__class__.OP_AND)
        if '' in lefts or left and not all(lefts):
            raise GraphParseError(
                f"Null task name in graph: {left} => {right}")

        for left in lefts:
            # Extract information about all nodes on the left.

            if left:
                info = self.__class__.REC_NODES.findall(left)
                expr = left

            else:
                # There is no left-hand-side task.
                info = []
                expr = ''

            n_info: List[Tuple[str, str, str, bool]] = []
            for name, offset, trig, opt_char in info:
                opt = opt_char == self.__class__.OPTIONAL
                if name.startswith(self.__class__.XTRIG):
                    n_info.append((name, offset, trig, opt))
                    continue
                if trig:
                    # Replace with standard trigger name if necessary
                    trig = trig.strip(self.__class__.QUALIFIER)
                    n_trig = TaskTrigger.standardise_name(trig)
                    if n_trig != trig:
                        if offset:
                            this = r'\b%s\b%s:%s(?!:)' % (
                                re.escape(name),
                                re.escape(offset),
                                re.escape(trig)
                            )
                        else:
                            this = r'\b%s:%s\b(?![\[:])' % (
                                re.escape(name),
                                re.escape(trig)
                            )
                        that = f"{name}{offset}:{n_trig}"
                        expr = re.sub(this, that, expr)
                else:
                    # Make success triggers explicit.
                    n_trig = TASK_OUTPUT_SUCCEEDED
                    if offset:
                        this = r'\b%s\b%s(?!:)' % (
                            re.escape(name),
                            re.escape(offset)
                        )
                    else:
                        this = r'\b%s\b(?![\[:])' % re.escape(name)
                    that = f"{name}{offset}:{n_trig}"
                    expr = re.sub(this, that, expr)

                n_info.append((name, offset, n_trig, opt))

            info = n_info

            # Determine semantics of all family triggers present.
            family_trig_map = {}
            for name, _, trig, _ in info:
                if name.startswith(self.__class__.XTRIG):
                    # Avoid @xtrigger nodes.
                    continue
                if name in self.family_map:
                    # Family; deal with members.
                    try:
                        family_trig_map[(name, trig)] = (
                            self.__class__.fam_to_mem_trigger_map[trig]
                        )
                    except KeyError:
                        # "FAM:bad => foo" in LHS (includes "FAM => bar" too).
                        raise GraphParseError(
                            f"Illegal family trigger in {expr}")
                else:
                    # Not a family.
                    if trig in self.__class__.fam_to_mem_trigger_map:
                        raise GraphParseError(
                            "family trigger on non-family namespace {expr}")

            # remove '?' from expr (not needed in logical trigger evaluation)
            expr = re.sub(self.__class__._RE_OPT, '', expr)
            self._families_all_to_all(expr, rights, info, family_trig_map)

    def _families_all_to_all(
        self,
        expr: str,
        rights: List[str],
        info: List[Tuple[str, str, str, bool]],
        family_trig_map: Dict[Tuple[str, str], Tuple[str, bool]]
    ) -> None:
        """Replace all family names with member names, for all/any semantics.

        Args:
            expr: the associated graph expression
            rights: list of right-side nodes
            trigs: parsed trigger info
            info: [(name, offset, trigger-name, optional)] for each node
            expr: the associated graph expression for this graph line

        """
        # Process left-side expression for defining triggers.
        n_info = []
        n_expr = expr
        for name, offset, trig, _ in info:
            if (name, trig) in family_trig_map:
                ttype, mem_all = family_trig_map[(name, trig)]
                m_info = []
                m_expr = []
                for mem in self.family_map[name]:
                    m_info.append((mem, offset, ttype))
                    m_expr.append(f"{mem}{offset}:{ttype}")
                this = r'\b%s%s:%s\b' % (
                    name,
                    re.escape(offset),
                    trig
                )
                if mem_all:
                    that = '(%s)' % '&'.join(m_expr)
                else:
                    that = '(%s)' % '|'.join(m_expr)
                n_expr = re.sub(this, that, n_expr)
                n_info += m_info
            else:
                n_info += [(name, offset, trig)]

        self._compute_triggers(expr, rights, n_expr, n_info)

    def _set_triggers(
        self,
        name: str,
        suicide: bool,
        trigs: List[str],
        expr: str,
        orig_expr: str
    ) -> None:
        """Record parsed triggers.

        Args:
            name: task name
            suicide: whether this is a suicide trigger or not
            trigs: parsed trigger info
            expr: the associated graph expression
            orig_expr: the original associated graph expression
        """

        # Check suicide triggers
        with contextlib.suppress(KeyError):
            osuicide = self.triggers[name][expr][1]
            # This trigger already exists, so we must have both
            # "expr => member" and "expr => !member" in the graph,
            # or simply a duplicate trigger not recognized earlier
            # because of parameter offsets.
            if not expr:
                pass
            elif suicide is not osuicide:
                oexp = re.sub(r'(&|\|)', r' \1 ', orig_expr)
                oexp = re.sub(r':succeeded', '', oexp)
                raise GraphParseError(
                    f"{oexp} can't trigger both {name} and !{name}")

        # Record triggers
        self.triggers.setdefault(name, {})
        self.triggers[name][expr] = (trigs, suicide)
        self.original.setdefault(name, {})
        self.original[name][expr] = orig_expr

    def _set_output_opt(
        self,
        name: str,
        output: str,
        optional: bool,
        suicide: bool,
        fam_member: bool = False
    ) -> None:
        """Set or check consistency of optional/required output.

        Args:
            name: task name
            output: task output name
            optional: is the output optional?
            suicide: is this from a suicide trigger?
            fam_member: is this from an expanded family trigger?

        """
        if cylc.flow.flags.cylc7_back_compat:
            # Set all outputs optional (set :succeed required elsewhere).
            self.task_output_opt[(name, output)] = (True, True, True)
            return

        # Do not infer output optionality from suicide triggers:
        if suicide:
            return

        if (
            output in {TASK_OUTPUT_EXPIRED, TASK_OUTPUT_SUBMIT_FAILED}
            and not optional
        ):
            # ":expire" and ":submit-fail" cannot be required
            # proposal point 4:
            # https://cylc.github.io/cylc-admin/proposal-optional-output-extension.html#proposal
            raise GraphParseError(f"{name}:{output} must be optional")

        if output == TASK_OUTPUT_FINISHED:
            # Interpret :finish pseudo-output
            if optional:
                raise GraphParseError(
                    f"Pseudo-output {name}:{output} can't be optional")
            # But implicit optional for the real succeed/fail outputs.
            optional = True
            for outp in [TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED]:
                self._set_output_opt(
                    name, outp, optional, suicide, fam_member)

        try:
            prev_optional, prev_default, prev_fixed = (
                self.task_output_opt[(name, output)])
        except KeyError:
            # Not already set; set it. Fix it if not fam_member.
            self.task_output_opt[(name, output)] = (
                optional, optional, not fam_member)
        else:
            # Already set; check consistency with previous.
            if prev_fixed:
                # optionality fixed already
                if fam_member:
                    pass
                else:
                    if optional != prev_optional:
                        raise GraphParseError(
                            f"Output {name}:{output} can't be both required"
                            " and optional")
            else:
                # optionality not fixed yet (only family default)
                if fam_member:
                    # family defaults must be consistent
                    if optional != prev_default:
                        raise GraphParseError(
                            f"Output {name}:{output} can't default to both"
                            " optional and required (via family trigger"
                            " defaults)")
                else:
                    # fix the optionality now
                    self.task_output_opt[(name, output)] = (
                        optional, prev_default, True)

        # Check opposite output where appropriate.
        for opposites in [
            (TASK_OUTPUT_SUCCEEDED, TASK_OUTPUT_FAILED),
            (TASK_OUTPUT_SUBMITTED, TASK_OUTPUT_SUBMIT_FAILED)
        ]:
            if output not in opposites:
                continue
            succeed, fail = opposites
            opposite = fail if output == succeed else succeed
            try:
                opp_optional, opp_default, opp_fixed = (
                    self.task_output_opt[(name, opposite)]
                )
            except KeyError:
                # opposite not set, no need to check
                continue
            else:
                # opposite already set; check consistency
                optional, default, oset = (
                    self.task_output_opt[(name, output)]
                )
                msg = (f"Opposite outputs {name}:{output} and {name}:"
                       f"{opposite} must both be optional if both are used")
                if fam_member or not opp_fixed:
                    if not optional or not opp_default:
                        raise GraphParseError(
                            msg + " (via family trigger defaults)")
                    elif not optional or not opp_optional:
                        raise GraphParseError(
                            msg + " (via family trigger)")
                elif not optional or not opp_optional:
                    raise GraphParseError(msg)

    def _compute_triggers(
        self,
        orig_expr: str,
        rights: List[str],
        expr: str,
        info: List[Tuple[str, str, str]]
    ) -> None:
        """Store trigger info from "expr => right".

        Args:
            orig_expr: the original associated graph expression
            rights: list of right-side nodes including qualifiers like :fail?
            expr: the associated graph expression
            info: [(name, offset, trigger-name)] for each name in expr.

        """
        trigs = []
        for name, offset, trigger in info:
            # Replace finish triggers (must be done after member substn).
            if name.startswith(self.__class__.XTRIG):
                trigs += [name]
            elif trigger == TASK_OUTPUT_FINISHED:
                this = f"{name}{offset}:{trigger}"
                that = "(%s%s:%s%s%s%s:%s)" % (
                    name, offset, TASK_OUTPUT_SUCCEEDED,
                    self.__class__.OP_OR,
                    name, offset, TASK_OUTPUT_FAILED)
                expr = expr.replace(this, that)
                trigs += [
                    "%s%s:%s" % (name, offset, TASK_OUTPUT_SUCCEEDED),
                    "%s%s:%s" % (name, offset, TASK_OUTPUT_FAILED)]
            else:
                trigs += [f"{name}{offset}:{trigger}"]

        for right in rights:
            right = right.strip('()')  # parentheses don't matter
            m = self.__class__.REC_RHS_NODE.match(right)
            if not m:
                # Bad nodes should have been detected earlier; fail loudly
                raise ValueError(  # pragma: no cover
                    f"Unexpected graph expression: '{right}'"
                )
            suicide_char, name, output, opt_char = m.groups()
            suicide = (suicide_char == self.__class__.SUICIDE)
            optional = (opt_char == self.__class__.OPTIONAL)
            if output:
                output = output.strip(self.__class__.QUALIFIER)

            if name in self.family_map:
                fam = True
                mems = self.family_map[name]
                if not output:
                    # (Plain family name on RHS).
                    # Make implicit success explicit.
                    output = QUAL_FAM_SUCCEED_ALL
                elif output.startswith("finish"):
                    if optional:
                        raise GraphParseError(
                            f"Family pseudo-output {name}:{output} can't be"
                            " optional")
                    # But implicit optional for the real succeed/fail outputs.
                    optional = True
                try:
                    outputs = self.__class__.fam_to_mem_output_map[output]
                except KeyError:
                    # Illegal family trigger on RHS of a pair.
                    raise GraphParseError(
                        f"Illegal family trigger: {name}:{output}")
            else:
                fam = False
                if not output:
                    # Make implicit success explicit.
                    output = TASK_OUTPUT_SUCCEEDED
                else:
                    # Convert to standard output names if necessary.
                    output = TaskTrigger.standardise_name(output)
                mems = [name]
                outputs = [output]

            for mem in mems:
                self._set_triggers(mem, suicide, trigs, expr, orig_expr)
                for output in outputs:
                    self._set_output_opt(mem, output, optional, suicide, fam)
